package javax.swing.text;

import java.beans.PropertyChangeEvent;
import javax.swing.event.DocumentEvent;
import java.awt.event.ActionListener;
import javax.swing.event.DocumentListener;
import java.beans.PropertyChangeListener;
import java.io.ObjectOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.awt.datatransfer.ClipboardOwner;
import java.awt.datatransfer.StringSelection;
import javax.swing.JPasswordField;
import javax.swing.plaf.TextUI;
import java.util.EventListener;
import javax.swing.event.ChangeListener;
import java.awt.Graphics;
import java.awt.datatransfer.Transferable;
import javax.swing.TransferHandler;
import java.awt.datatransfer.Clipboard;
import java.awt.HeadlessException;
import javax.swing.UIManager;
import sun.swing.SwingUtilities2;
import javax.swing.ActionMap;
import java.awt.event.ActionEvent;
import java.awt.event.FocusEvent;
import javax.swing.SwingUtilities;
import java.awt.event.MouseEvent;
import javax.swing.Action;
import java.awt.Point;
import javax.swing.Timer;
import javax.swing.event.ChangeEvent;
import javax.swing.event.EventListenerList;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseListener;
import java.awt.event.FocusListener;
import java.awt.Rectangle;

public class DefaultCaret extends Rectangle implements Caret, FocusListener, MouseListener, MouseMotionListener
{
    public static final int UPDATE_WHEN_ON_EDT = 0;
    public static final int NEVER_UPDATE = 1;
    public static final int ALWAYS_UPDATE = 2;
    protected EventListenerList listenerList;
    protected transient ChangeEvent changeEvent;
    JTextComponent component;
    int updatePolicy;
    boolean visible;
    boolean active;
    int dot;
    int mark;
    Object selectionTag;
    boolean selectionVisible;
    Timer flasher;
    Point magicCaretPosition;
    transient Position.Bias dotBias;
    transient Position.Bias markBias;
    boolean dotLTR;
    boolean markLTR;
    transient DefaultCaret.Handler handler;
    private transient int[] flagXPoints;
    private transient int[] flagYPoints;
    private transient NavigationFilter.FilterBypass filterBypass;
    private static transient Action selectWord;
    private static transient Action selectLine;
    private boolean ownsSelection;
    private boolean forceCaretPositionChange;
    private transient boolean shouldHandleRelease;
    private transient MouseEvent selectedWordEvent;
    private int caretWidth;
    private float aspectRatio;
    
    public DefaultCaret() {
        this.listenerList = new EventListenerList();
        this.changeEvent = null;
        this.updatePolicy = 0;
        this.handler = new DefaultCaret.Handler();
        this.flagXPoints = new int[3];
        this.flagYPoints = new int[3];
        this.selectedWordEvent = null;
        this.caretWidth = -1;
        this.aspectRatio = -1.0f;
    }
    
    public void setUpdatePolicy(int updatePolicy) {
        this.updatePolicy = updatePolicy;
    }
    
    public int getUpdatePolicy() {
        return this.updatePolicy;
    }
    
    protected final JTextComponent getComponent() {
        return this.component;
    }
    
    protected final synchronized void repaint() {
        if (this.component != null) {
            this.component.repaint(this.x, this.y, this.width, this.height);
        }
    }
    
    protected synchronized void damage(Rectangle rectangle) {
        if (rectangle != null) {
            int caretWidth = this.getCaretWidth(rectangle.height);
            this.x = rectangle.x - 4 - (caretWidth >> 1);
            this.y = rectangle.y;
            this.width = 9 + caretWidth;
            this.height = rectangle.height;
            this.repaint();
        }
    }
    
    protected void adjustVisibility(Rectangle aRect) {
        if (this.component == null) {
            return;
        }
        if (SwingUtilities.isEventDispatchThread()) {
            this.component.scrollRectToVisible(aRect);
        }
        else {
            SwingUtilities.invokeLater(new DefaultCaret.SafeScroller(aRect));
        }
        return;
    }
    
    protected Highlighter.HighlightPainter getSelectionPainter() {
        return DefaultHighlighter.DefaultPainter;
    }
    
    protected void positionCaret(MouseEvent mouseEvent) {
        Point point = new Point(mouseEvent.getX(), mouseEvent.getY());
        Position.Bias[] array = { null };
        int viewToModel = this.component.getUI().viewToModel(this.component, point, array);
        if (array[0] == null) {
            array[0] = Position.Bias.Forward;
        }
        if (viewToModel >= 0) {
            this.setDot(viewToModel, array[0]);
        }
    }
    
    protected void moveCaret(MouseEvent mouseEvent) {
        Point point = new Point(mouseEvent.getX(), mouseEvent.getY());
        Position.Bias[] array = { null };
        int viewToModel = this.component.getUI().viewToModel(this.component, point, array);
        if (array[0] == null) {
            array[0] = Position.Bias.Forward;
        }
        if (viewToModel >= 0) {
            this.moveDot(viewToModel, array[0]);
        }
    }
    
    @Override
    public void focusGained(FocusEvent focusEvent) {
        if (this.component.isEnabled()) {
            if (this.component.isEditable()) {
                this.setVisible(true);
            }
            this.setSelectionVisible(true);
        }
    }
    
    @Override
    public void focusLost(FocusEvent focusEvent) {
        this.setVisible(false);
        this.setSelectionVisible(this.ownsSelection || focusEvent.isTemporary());
    }
    
    private void selectWord(MouseEvent selectedWordEvent) {
        if (this.selectedWordEvent != null && this.selectedWordEvent.getX() == selectedWordEvent.getX() && this.selectedWordEvent.getY() == selectedWordEvent.getY()) {
            return;
        }
        Action action = null;
        ActionMap actionMap = this.getComponent().getActionMap();
        if (actionMap != null) {
            action = actionMap.get("select-word");
        }
        if (action == null) {
            if (DefaultCaret.selectWord == null) {
                DefaultCaret.selectWord = new DefaultEditorKit.SelectWordAction();
            }
            action = DefaultCaret.selectWord;
        }
        action.actionPerformed(new ActionEvent(this.getComponent(), 1001, null, selectedWordEvent.getWhen(), selectedWordEvent.getModifiers()));
        this.selectedWordEvent = selectedWordEvent;
        return;
    }
    
    @Override
    public void mouseClicked(MouseEvent mouseEvent) {
        if (this.getComponent() == null) {
            return;
        }
        int adjustedClickCount = SwingUtilities2.getAdjustedClickCount(this.getComponent(), mouseEvent);
        if (!mouseEvent.isConsumed()) {
            if (SwingUtilities.isLeftMouseButton(mouseEvent)) {
                if (adjustedClickCount == 1) {
                    this.selectedWordEvent = null;
                }
                else if (adjustedClickCount == 2 && SwingUtilities2.canEventAccessSystemClipboard(mouseEvent)) {
                    this.selectWord(mouseEvent);
                    this.selectedWordEvent = null;
                }
                else if (adjustedClickCount == 3 && SwingUtilities2.canEventAccessSystemClipboard(mouseEvent)) {
                    Action action = null;
                    ActionMap actionMap = this.getComponent().getActionMap();
                    if (actionMap != null) {
                        action = actionMap.get("select-line");
                    }
                    if (action == null) {
                        if (DefaultCaret.selectLine == null) {
                            DefaultCaret.selectLine = new DefaultEditorKit.SelectLineAction();
                        }
                        action = DefaultCaret.selectLine;
                    }
                    action.actionPerformed(new ActionEvent(this.getComponent(), 1001, null, mouseEvent.getWhen(), mouseEvent.getModifiers()));
                }
            }
            else if (SwingUtilities.isMiddleMouseButton(mouseEvent) && adjustedClickCount == 1 && this.component.isEditable() && this.component.isEnabled() && SwingUtilities2.canEventAccessSystemClipboard(mouseEvent)) {
                JTextComponent textComponent = (JTextComponent)mouseEvent.getSource();
                if (textComponent != null) {
                    try {
                        Clipboard systemSelection = textComponent.getToolkit().getSystemSelection();
                        if (systemSelection != null) {
                            this.adjustCaret(mouseEvent);
                            TransferHandler transferHandler = textComponent.getTransferHandler();
                            if (transferHandler != null) {
                                Transferable contents = null;
                                try {
                                    contents = systemSelection.getContents(null);
                                }
                                catch (IllegalStateException ex) {
                                    UIManager.getLookAndFeel().provideErrorFeedback(textComponent);
                                }
                                if (contents != null) {
                                    transferHandler.importData(textComponent, contents);
                                }
                            }
                            this.adjustFocus(true);
                        }
                    }
                    catch (HeadlessException ex2) {}
                }
            }
        }
        return;
    }
    
    @Override
    public void mousePressed(MouseEvent e) {
        int adjustedClickCount = SwingUtilities2.getAdjustedClickCount(this.getComponent(), e);
        if (SwingUtilities.isLeftMouseButton(e)) {
            if (e.isConsumed()) {
                this.shouldHandleRelease = true;
            }
            else {
                this.shouldHandleRelease = false;
                this.adjustCaretAndFocus(e);
                if (adjustedClickCount == 2 && SwingUtilities2.canEventAccessSystemClipboard(e)) {
                    this.selectWord(e);
                }
            }
        }
    }
    
    void adjustCaretAndFocus(MouseEvent mouseEvent) {
        this.adjustCaret(mouseEvent);
        this.adjustFocus(false);
    }
    
    private void adjustCaret(MouseEvent mouseEvent) {
        if ((mouseEvent.getModifiers() & 0x1) != 0x0 && this.getDot() != -1) {
            this.moveCaret(mouseEvent);
        }
        else if (!mouseEvent.isPopupTrigger()) {
            this.positionCaret(mouseEvent);
        }
    }
    
    private void adjustFocus(boolean b) {
        if (this.component != null && this.component.isEnabled() && this.component.isRequestFocusEnabled()) {
            if (b) {
                this.component.requestFocusInWindow();
            }
            else {
                this.component.requestFocus();
            }
        }
    }
    
    @Override
    public void mouseReleased(MouseEvent anEvent) {
        if (!anEvent.isConsumed() && this.shouldHandleRelease && SwingUtilities.isLeftMouseButton(anEvent)) {
            this.adjustCaretAndFocus(anEvent);
        }
    }
    
    @Override
    public void mouseEntered(MouseEvent mouseEvent) {
    }
    
    @Override
    public void mouseExited(MouseEvent mouseEvent) {
    }
    
    @Override
    public void mouseDragged(MouseEvent anEvent) {
        if (!anEvent.isConsumed() && SwingUtilities.isLeftMouseButton(anEvent)) {
            this.moveCaret(anEvent);
        }
    }
    
    @Override
    public void mouseMoved(MouseEvent mouseEvent) {
    }
    
    @Override
    public void paint(Graphics graphics) {
        if (this.isVisible()) {
            try {
                Rectangle modelToView = this.component.getUI().modelToView(this.component, this.dot, this.dotBias);
                if (modelToView == null || (modelToView.width == 0 && modelToView.height == 0)) {
                    return;
                }
                if (this.width > 0 && this.height > 0 && !this._contains(modelToView.x, modelToView.y, modelToView.width, modelToView.height)) {
                    Rectangle clipBounds = graphics.getClipBounds();
                    if (clipBounds != null && !clipBounds.contains(this)) {
                        this.repaint();
                    }
                    this.damage(modelToView);
                }
                graphics.setColor(this.component.getCaretColor());
                int caretWidth = this.getCaretWidth(modelToView.height);
                graphics.fillRect(modelToView.x -= caretWidth >> 1, modelToView.y, caretWidth, modelToView.height);
                Document document = this.component.getDocument();
                if ((document instanceof AbstractDocument)) {
                    Element bidiRootElement = ((AbstractDocument)document).getBidiRootElement();
                    if (bidiRootElement != null && bidiRootElement.getElementCount() > 1) {
                        this.flagXPoints[0] = modelToView.x + (this.dotLTR ? caretWidth : false);
                        this.flagYPoints[0] = modelToView.y;
                        this.flagXPoints[1] = this.flagXPoints[0];
                        this.flagYPoints[1] = this.flagYPoints[0] + 4;
                        this.flagXPoints[2] = this.flagXPoints[0] + (this.dotLTR ? 4 : -4);
                        this.flagYPoints[2] = this.flagYPoints[0];
                        graphics.fillPolygon(this.flagXPoints, this.flagYPoints, 3);
                    }
                }
            }
            catch (BadLocationException ex) {}
        }
    }
    
    @Override
    public void install(JTextComponent component) {
        this.component = component;
        Document document = component.getDocument();
        int n = 0;
        this.mark = n;
        this.dot = n;
        boolean b = true;
        this.markLTR = b;
        this.dotLTR = b;
        Position.Bias forward = Position.Bias.Forward;
        this.markBias = forward;
        this.dotBias = forward;
        if (document != null) {
            document.addDocumentListener(this.handler);
        }
        component.addPropertyChangeListener(this.handler);
        component.addFocusListener(this);
        component.addMouseListener(this);
        component.addMouseMotionListener(this);
        if (this.component.hasFocus()) {
            this.focusGained(null);
        }
        Number n2 = (Number)component.getClientProperty("caretAspectRatio");
        if (n2 != null) {
            this.aspectRatio = n2.floatValue();
        }
        else {
            this.aspectRatio = -1.0f;
        }
        Integer n3 = (Integer)component.getClientProperty("caretWidth");
        if (n3 != null) {
            this.caretWidth = (int)n3;
        }
        else {
            this.caretWidth = -1;
        }
    }
    
    @Override
    public void deinstall(JTextComponent textComponent) {
        textComponent.removeMouseListener(this);
        textComponent.removeMouseMotionListener(this);
        textComponent.removeFocusListener(this);
        textComponent.removePropertyChangeListener(this.handler);
        Document document = textComponent.getDocument();
        if (document != null) {
            document.removeDocumentListener(this.handler);
        }
        synchronized (this) {
            this.component = null;
        }
        if (this.flasher != null) {
            this.flasher.stop();
        }
    }
    
    @Override
    public void addChangeListener(ChangeListener l) {
        this.listenerList.add(ChangeListener.class, l);
    }
    
    @Override
    public void removeChangeListener(ChangeListener l) {
        this.listenerList.remove(ChangeListener.class, l);
    }
    
    public ChangeListener[] getChangeListeners() {
        return (ChangeListener[])this.listenerList.getListeners(ChangeListener.class);
    }
    
    protected void fireStateChanged() {
        Object[] listenerList = this.listenerList.getListenerList();
        for (int i = listenerList.length - 2; i >= 0; i -= 2) {
            if (listenerList[i] == ChangeListener.class) {
                if (this.changeEvent == null) {
                    this.changeEvent = new ChangeEvent(this);
                }
                ((ChangeListener)listenerList[i + 1]).stateChanged(this.changeEvent);
            }
        }
    }
    
    public <T extends EventListener> T[] getListeners(Class<T> t) {
        return this.listenerList.getListeners(t);
    }
    
    @Override
    public void setSelectionVisible(boolean selectionVisible) {
        if (selectionVisible != this.selectionVisible) {
            this.selectionVisible = selectionVisible;
            if (this.selectionVisible) {
                Highlighter highlighter = this.component.getHighlighter();
                if (this.dot != this.mark && highlighter != null && this.selectionTag == null) {
                    int min = Math.min(this.dot, this.mark);
                    int max = Math.max(this.dot, this.mark);
                    Highlighter.HighlightPainter selectionPainter = this.getSelectionPainter();
                    try {
                        this.selectionTag = highlighter.addHighlight(min, max, selectionPainter);
                    }
                    catch (BadLocationException ex) {
                        this.selectionTag = null;
                    }
                }
            }
            else if (this.selectionTag != null) {
                this.component.getHighlighter().removeHighlight(this.selectionTag);
                this.selectionTag = null;
            }
        }
    }
    
    @Override
    public boolean isSelectionVisible() {
        return this.selectionVisible;
    }
    
    public boolean isActive() {
        return this.active;
    }
    
    @Override
    public boolean isVisible() {
        return this.visible;
    }
    
    @Override
    public void setVisible(boolean b) {
        this.active = b;
        if (this.component != null) {
            TextUI ui = this.component.getUI();
            if (this.visible != b) {
                this.visible = b;
                try {
                    this.damage(ui.modelToView(this.component, this.dot, this.dotBias));
                }
                catch (BadLocationException ex) {}
            }
        }
        if (this.flasher != null) {
            if (this.visible) {
                this.flasher.start();
            }
            else {
                this.flasher.stop();
            }
        }
    }
    
    @Override
    public void setBlinkRate(int n) {
        if (n != 0) {
            if (this.flasher == null) {
                this.flasher = new Timer(n, this.handler);
            }
            this.flasher.setDelay(n);
        }
        else if (this.flasher != null) {
            this.flasher.stop();
            this.flasher.removeActionListener(this.handler);
            this.flasher = null;
        }
    }
    
    @Override
    public int getBlinkRate() {
        return (this.flasher == null) ? 0 : this.flasher.getDelay();
    }
    
    @Override
    public int getDot() {
        return this.dot;
    }
    
    @Override
    public int getMark() {
        return this.mark;
    }
    
    @Override
    public void setDot(int n) {
        this.setDot(n, Position.Bias.Forward);
    }
    
    @Override
    public void moveDot(int n) {
        this.moveDot(n, Position.Bias.Forward);
    }
    
    public void moveDot(int n, Position.Bias bias) {
        if (bias == null) {
            throw new IllegalArgumentException("null bias");
        }
        if (!this.component.isEnabled()) {
            this.setDot(n, bias);
            return;
        }
        if (n != this.dot) {
            NavigationFilter navigationFilter = this.component.getNavigationFilter();
            if (navigationFilter != null) {
                navigationFilter.moveDot(this.getFilterBypass(), n, bias);
            }
            else {
                this.handleMoveDot(n, bias);
            }
        }
        return;
    }
    
    void handleMoveDot(int n, Position.Bias bias) {
        this.changeCaretPosition(n, bias);
        if (this.selectionVisible) {
            Highlighter highlighter = this.component.getHighlighter();
            if (highlighter != null) {
                int min = Math.min(n, this.mark);
                int max = Math.max(n, this.mark);
                if (min == max) {
                    if (this.selectionTag != null) {
                        highlighter.removeHighlight(this.selectionTag);
                        this.selectionTag = null;
                    }
                }
                else {
                    try {
                        if (this.selectionTag != null) {
                            highlighter.changeHighlight(this.selectionTag, min, max);
                        }
                        else {
                            this.selectionTag = highlighter.addHighlight(min, max, this.getSelectionPainter());
                        }
                    }
                    catch (BadLocationException ex) {
                        throw new StateInvariantError("Bad caret position");
                    }
                }
            }
        }
    }
    
    public void setDot(int n, Position.Bias bias) {
        if (bias == null) {
            throw new IllegalArgumentException("null bias");
        }
        NavigationFilter navigationFilter = this.component.getNavigationFilter();
        if (navigationFilter != null) {
            navigationFilter.setDot(this.getFilterBypass(), n, bias);
        }
        else {
            this.handleSetDot(n, bias);
        }
        return;
    }
    
    void handleSetDot(int mark, Position.Bias forward) {
        Document document = this.component.getDocument();
        if (document != null) {
            mark = Math.min(mark, document.getLength());
        }
        mark = Math.max(mark, 0);
        if (mark == 0) {
            forward = Position.Bias.Forward;
        }
        this.mark = mark;
        if (this.dot != mark || this.dotBias != forward || this.selectionTag != null || this.forceCaretPositionChange) {
            this.changeCaretPosition(mark, forward);
        }
        this.markBias = this.dotBias;
        this.markLTR = this.dotLTR;
        Highlighter highlighter = this.component.getHighlighter();
        if (highlighter != null && this.selectionTag != null) {
            highlighter.removeHighlight(this.selectionTag);
            this.selectionTag = null;
        }
    }
    
    public Position.Bias getDotBias() {
        return this.dotBias;
    }
    
    public Position.Bias getMarkBias() {
        return this.markBias;
    }
    
    boolean isDotLeftToRight() {
        return this.dotLTR;
    }
    
    boolean isMarkLeftToRight() {
        return this.markLTR;
    }
    
    boolean isPositionLTR(int n, Position.Bias bias) {
        Document document = this.component.getDocument();
        if (bias == Position.Bias.Backward && --n < 0) {
            n = 0;
        }
        return AbstractDocument.isLeftToRight(document, n, n);
    }
    
    Position.Bias guessBiasForOffset(int n, Position.Bias bias, boolean b) {
        if (b != this.isPositionLTR(n, bias)) {
            bias = Position.Bias.Backward;
        }
        else if (bias != Position.Bias.Backward && b != this.isPositionLTR(n, Position.Bias.Backward)) {
            bias = Position.Bias.Backward;
        }
        if (bias == Position.Bias.Backward && n > 0) {
            try {
                Segment o = new Segment();
                this.component.getDocument().getText(n - 1, 1, o);
                if (o.count > 0 && o.array[o.offset] == '\n') {
                    bias = Position.Bias.Forward;
                }
            }
            catch (BadLocationException ex) {}
        }
        return bias;
    }
    
    void changeCaretPosition(int dot, Position.Bias dotBias) {
        this.repaint();
        if (this.flasher != null && this.flasher.isRunning()) {
            this.visible = true;
            this.flasher.restart();
        }
        this.dot = dot;
        this.dotBias = dotBias;
        this.dotLTR = this.isPositionLTR(dot, dotBias);
        this.fireStateChanged();
        this.updateSystemSelection();
        this.setMagicCaretPosition(null);
        SwingUtilities.invokeLater(new Runnable() {
            @Override
            public void run() {
                DefaultCaret.this.repaintNewCaret();
            }
        });
    }
    
    void repaintNewCaret() {
        if (this.component != null) {
            TextUI ui = this.component.getUI();
            Document document = this.component.getDocument();
            if (ui != null && document != null) {
                Rectangle modelToView;
                try {
                    modelToView = ui.modelToView(this.component, this.dot, this.dotBias);
                }
                catch (BadLocationException ex) {
                    modelToView = null;
                }
                if (modelToView != null) {
                    this.adjustVisibility(modelToView);
                    if (this.getMagicCaretPosition() == null) {
                        this.setMagicCaretPosition(new Point(modelToView.x, modelToView.y));
                    }
                }
                this.damage(modelToView);
            }
        }
    }
    
    private void updateSystemSelection() {
        if (!SwingUtilities2.canCurrentEventAccessSystemClipboard()) {
            return;
        }
        if (this.dot != this.mark && this.component != null && this.component.hasFocus()) {
            Clipboard systemSelection = this.getSystemSelection();
            if (systemSelection != null) {
                String selectedText;
                if (this.component instanceof JPasswordField && this.component.getClientProperty("JPasswordField.cutCopyAllowed") != Boolean.TRUE) {
                    Object o = null;
                    int echoChar = ((JPasswordField)this.component).getEchoChar();
                    int i;
                    int max;
                    for (i = Math.min(this.getDot(), this.getMark()), max = Math.max(this.getDot(), this.getMark()), ; i < max; ++i) {
                        if (o == null) {
                            o = new StringBuilder();
                        }
                        o.append((char)echoChar);
                    }
                    selectedText = ((o != null) ? o.toString() : null);
                }
                else {
                    selectedText = this.component.getSelectedText();
                }
                try {
                    systemSelection.setContents(new StringSelection(selectedText), this.getClipboardOwner());
                    this.ownsSelection = true;
                }
                catch (IllegalStateException ex) {}
            }
        }
        return;
    }
    
    private Clipboard getSystemSelection() {
        try {
            return this.component.getToolkit().getSystemSelection();
        }
        catch (HeadlessException ex) {}
        catch (SecurityException ex2) {}
        return null;
    }
    
    private ClipboardOwner getClipboardOwner() {
        return this.handler;
    }
    
    private void ensureValidPosition() {
        int length = this.component.getDocument().getLength();
        if (this.dot > length || this.mark > length) {
            this.handleSetDot(length, Position.Bias.Forward);
        }
    }
    
    @Override
    public void setMagicCaretPosition(Point magicCaretPosition) {
        this.magicCaretPosition = magicCaretPosition;
    }
    
    @Override
    public Point getMagicCaretPosition() {
        return this.magicCaretPosition;
    }
    
    @Override
    public boolean equals(Object o) {
        return this == o;
    }
    
    @Override
    public String toString() {
        return "Dot=(" + this.dot + ", " + this.dotBias + ")" + " Mark=(" + this.mark + ", " + this.markBias + ")";
    }
    
    private NavigationFilter.FilterBypass getFilterBypass() {
        if (this.filterBypass == null) {
            this.filterBypass = new DefaultCaret.DefaultFilterBypass();
        }
        return this.filterBypass;
    }
    
    private boolean _contains(int n, int n2, int n3, int n4) {
        int width = this.width;
        int height = this.height;
        if ((width | height | n3 | n4) < 0) {
            return false;
        }
        int x = this.x;
        int y = this.y;
        if (n < x || n2 < y) {
            return false;
        }
        if (n3 > 0) {
            int n5 = width + x;
            n3 += n;
            if (n3 <= n) {
                if (n5 >= x || n3 > n5) {
                    return false;
                }
            }
            else if (n5 >= x && n3 > n5) {
                return false;
            }
        }
        else if (x + width < n) {
            return false;
        }
        if (n4 > 0) {
            int n6 = height + y;
            n4 += n2;
            if (n4 <= n2) {
                if (n6 >= y || n4 > n6) {
                    return false;
                }
            }
            else if (n6 >= y && n4 > n6) {
                return false;
            }
        }
        else if (y + height < n2) {
            return false;
        }
        return true;
    }
    
    int getCaretWidth(int n) {
        if (this.aspectRatio > -1.0f) {
            return (int)(this.aspectRatio * (float)n) + 1;
        }
        if (this.caretWidth > -1) {
            return this.caretWidth;
        }
        Object value = UIManager.get("Caret.width");
        if ((value instanceof Integer)) {
            return (int)value;
        }
        return 1;
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws ClassNotFoundException, IOException {
        objectInputStream.defaultReadObject();
        this.handler = new DefaultCaret.Handler();
        if (!objectInputStream.readBoolean()) {
            this.dotBias = Position.Bias.Forward;
        }
        else {
            this.dotBias = Position.Bias.Backward;
        }
        if (!objectInputStream.readBoolean()) {
            this.markBias = Position.Bias.Forward;
        }
        else {
            this.markBias = Position.Bias.Backward;
        }
    }
    
    private void writeObject(ObjectOutputStream objectOutputStream) throws IOException {
        objectOutputStream.defaultWriteObject();
        objectOutputStream.writeBoolean(this.dotBias == Position.Bias.Backward);
        objectOutputStream.writeBoolean(this.markBias == Position.Bias.Backward);
    }
    
    static void access$100(DefaultCaret defaultCaret) {
        defaultCaret.ensureValidPosition();
    }
    
    static {
        DefaultCaret.selectWord = null;
        DefaultCaret.selectLine = null;
    }
    
    class SafeScroller implements Runnable
    {
        Rectangle r;
        
        SafeScroller(Rectangle r) {
            this.r = r;
        }
        
        @Override
        public void run() {
            if (DefaultCaret.this.component != null) {
                DefaultCaret.this.component.scrollRectToVisible(this.r);
            }
        }
    }
    
    class Handler implements PropertyChangeListener, DocumentListener, ActionListener, ClipboardOwner
    {
        @Override
        public void actionPerformed(ActionEvent actionEvent) {
            if ((DefaultCaret.this.width == 0 || DefaultCaret.this.height == 0) && DefaultCaret.this.component != null) {
                TextUI ui = DefaultCaret.this.component.getUI();
                try {
                    Rectangle modelToView = ui.modelToView(DefaultCaret.this.component, DefaultCaret.this.dot, DefaultCaret.this.dotBias);
                    if (modelToView != null && modelToView.width != 0 && modelToView.height != 0) {
                        DefaultCaret.this.damage(modelToView);
                    }
                }
                catch (BadLocationException ex) {}
            }
            DefaultCaret.this.visible = !DefaultCaret.this.visible;
            DefaultCaret.this.repaint();
        }
        
        @Override
        public void insertUpdate(DocumentEvent documentEvent) {
            if (DefaultCaret.this.getUpdatePolicy() == 1 || (DefaultCaret.this.getUpdatePolicy() == 0 && !SwingUtilities.isEventDispatchThread())) {
                if ((documentEvent.getOffset() <= DefaultCaret.this.dot || documentEvent.getOffset() <= DefaultCaret.this.mark) && DefaultCaret.this.selectionTag != null) {
                    try {
                        DefaultCaret.this.component.getHighlighter().changeHighlight(DefaultCaret.this.selectionTag, Math.min(DefaultCaret.this.dot, DefaultCaret.this.mark), Math.max(DefaultCaret.this.dot, DefaultCaret.this.mark));
                    }
                    catch (BadLocationException ex2) {
                        ex2.printStackTrace();
                    }
                }
                return;
            }
            else {
                int offset = documentEvent.getOffset();
                int length = documentEvent.getLength();
                int dot = DefaultCaret.this.dot;
                int n = 0;
                if ((documentEvent instanceof AbstractDocument.UndoRedoDocumentEvent)) {
                    DefaultCaret.this.setDot(offset + length);
                    return;
                }
                if (dot >= offset) {
                    dot += length;
                    n = (short)(n | 0x1);
                }
                int mark = DefaultCaret.this.mark;
                if (mark >= offset) {
                    mark += length;
                    n = (short)(n | 0x2);
                }
                if (n != 0) {
                    Position.Bias bias = DefaultCaret.this.dotBias;
                    if (DefaultCaret.this.dot == offset) {
                        Document document = DefaultCaret.this.component.getDocument();
                        int n2;
                        try {
                            Segment o = new Segment();
                            document.getText(dot - 1, 1, o);
                            n2 = (o.count > 0 && o.array[o.offset] == '\n');
                        }
                        catch (BadLocationException ex) {
                            n2 = false;
                        }
                        if (!(n2 == 0)) {
                            bias = Position.Bias.Forward;
                        }
                        else {
                            bias = Position.Bias.Backward;
                        }
                    }
                    if (mark == dot) {
                        DefaultCaret.this.setDot(dot, bias);
                        DefaultCaret.access$100(DefaultCaret.this);
                    }
                    else {
                        DefaultCaret.this.setDot(mark, DefaultCaret.this.markBias);
                        if (DefaultCaret.this.getDot() == mark) {
                            DefaultCaret.this.moveDot(dot, bias);
                        }
                        DefaultCaret.access$100(DefaultCaret.this);
                    }
                }
                return;
            }
        }
        
        @Override
        public void removeUpdate(DocumentEvent documentEvent) {
            if (DefaultCaret.this.getUpdatePolicy() == 1 || (DefaultCaret.this.getUpdatePolicy() == 0 && !SwingUtilities.isEventDispatchThread())) {
                int length = DefaultCaret.this.component.getDocument().getLength();
                DefaultCaret.this.dot = Math.min(DefaultCaret.this.dot, length);
                DefaultCaret.this.mark = Math.min(DefaultCaret.this.mark, length);
                if ((documentEvent.getOffset() < DefaultCaret.this.dot || documentEvent.getOffset() < DefaultCaret.this.mark) && DefaultCaret.this.selectionTag != null) {
                    try {
                        DefaultCaret.this.component.getHighlighter().changeHighlight(DefaultCaret.this.selectionTag, Math.min(DefaultCaret.this.dot, DefaultCaret.this.mark), Math.max(DefaultCaret.this.dot, DefaultCaret.this.mark));
                    }
                    catch (BadLocationException ex) {
                        ex.printStackTrace();
                    }
                }
                return;
            }
            else {
                int offset = documentEvent.getOffset();
                int n = offset + documentEvent.getLength();
                int dot = DefaultCaret.this.dot;
                boolean n2 = false;
                int mark = DefaultCaret.this.mark;
                boolean n3 = false;
                if ((documentEvent instanceof AbstractDocument.UndoRedoDocumentEvent)) {
                    DefaultCaret.this.setDot(offset);
                    return;
                }
                if (dot >= n) {
                    dot -= n - offset;
                    if (dot == n) {
                        n2 = true;
                    }
                }
                else if (dot >= offset) {
                    dot = offset;
                    n2 = true;
                }
                if (mark >= n) {
                    mark -= n - offset;
                    if (mark == n) {
                        n3 = true;
                    }
                }
                else if (mark >= offset) {
                    mark = offset;
                    n3 = true;
                }
                if (mark == dot) {
                    DefaultCaret.this.forceCaretPositionChange = true;
                    try {
                        DefaultCaret.this.setDot(dot, DefaultCaret.this.guessBiasForOffset(dot, DefaultCaret.this.dotBias, DefaultCaret.this.dotLTR));
                    }
                    finally {
                        DefaultCaret.this.forceCaretPositionChange = false;
                    }
                    DefaultCaret.access$100(DefaultCaret.this);
                }
                else {
                    Position.Bias bias = DefaultCaret.this.dotBias;
                    Position.Bias bias2 = DefaultCaret.this.markBias;
                    if (!(n2 == 0)) {
                        bias = DefaultCaret.this.guessBiasForOffset(dot, bias, DefaultCaret.this.dotLTR);
                    }
                    if (!(n3 == 0)) {
                        bias2 = DefaultCaret.this.guessBiasForOffset(DefaultCaret.this.mark, bias2, DefaultCaret.this.markLTR);
                    }
                    DefaultCaret.this.setDot(mark, bias2);
                    if (DefaultCaret.this.getDot() == mark) {
                        DefaultCaret.this.moveDot(dot, bias);
                    }
                    DefaultCaret.access$100(DefaultCaret.this);
                }
                return;
            }
        }
        
        @Override
        public void changedUpdate(DocumentEvent documentEvent) {
            if (DefaultCaret.this.getUpdatePolicy() == 1 || (DefaultCaret.this.getUpdatePolicy() == 0 && !SwingUtilities.isEventDispatchThread())) {
                return;
            }
            if ((documentEvent instanceof AbstractDocument.UndoRedoDocumentEvent)) {
                DefaultCaret.this.setDot(documentEvent.getOffset() + documentEvent.getLength());
            }
            return;
        }
        
        @Override
        public void propertyChange(PropertyChangeEvent propertyChangeEvent) {
            Object oldValue = propertyChangeEvent.getOldValue();
            Object newValue = propertyChangeEvent.getNewValue();
            if (oldValue instanceof Document || newValue instanceof Document) {
                DefaultCaret.this.setDot(0);
                if (oldValue != null) {
                    ((Document)oldValue).removeDocumentListener(this);
                }
                if (newValue != null) {
                    ((Document)newValue).addDocumentListener(this);
                }
            }
            else if ("enabled".equals(propertyChangeEvent.getPropertyName())) {
                Boolean b = (Boolean)propertyChangeEvent.getNewValue();
                if (DefaultCaret.this.component.isFocusOwner()) {
                    if (b == Boolean.TRUE) {
                        if (DefaultCaret.this.component.isEditable()) {
                            DefaultCaret.this.setVisible(true);
                        }
                        DefaultCaret.this.setSelectionVisible(true);
                    }
                    else {
                        DefaultCaret.this.setVisible(false);
                        DefaultCaret.this.setSelectionVisible(false);
                    }
                }
            }
            else if ("caretWidth".equals(propertyChangeEvent.getPropertyName())) {
                Integer n2 = (Integer)propertyChangeEvent.getNewValue();
                if (n2 != null) {
                    DefaultCaret.this.caretWidth = (int)n2;
                }
                else {
                    DefaultCaret.this.caretWidth = -1;
                }
                DefaultCaret.this.repaint();
            }
            else if ("caretAspectRatio".equals(propertyChangeEvent.getPropertyName())) {
                Number n = (Number)propertyChangeEvent.getNewValue();
                if (n != null) {
                    DefaultCaret.this.aspectRatio = n.floatValue();
                }
                else {
                    DefaultCaret.this.aspectRatio = -1.0f;
                }
                DefaultCaret.this.repaint();
            }
        }
        
        @Override
        public void lostOwnership(Clipboard clipboard, Transferable transferable) {
            if (DefaultCaret.this.ownsSelection) {
                DefaultCaret.this.ownsSelection = false;
                if (DefaultCaret.this.component != null && !DefaultCaret.this.component.hasFocus()) {
                    DefaultCaret.this.setSelectionVisible(false);
                }
            }
        }
    }
    
    private class DefaultFilterBypass extends NavigationFilter.FilterBypass
    {
        @Override
        public Caret getCaret() {
            return DefaultCaret.this;
        }
        
        @Override
        public void setDot(int n, Position.Bias bias) {
            DefaultCaret.this.handleSetDot(n, bias);
        }
        
        @Override
        public void moveDot(int n, Position.Bias bias) {
            DefaultCaret.this.handleMoveDot(n, bias);
        }
    }
}
